use std::sync::Arc;

use btleplug::{
    api::{Central, CentralState, Manager as _, Peripheral, ScanFilter},
    platform::{Adapter, Manager},
};
use futures::StreamExt;
use tauri::{
    async_runtime::{self, JoinHandle, RwLock},
    AppHandle, Emitter, State,
};

use crate::{cmd::PeripheralItem, errors, state::AppState};

pub async fn handle_bluetooth_events(
    app_handle: Arc<AppHandle>,
    app_state: Arc<RwLock<AppState>>,
) -> Result<JoinHandle<()>, errors::AppError> {
    let state = app_state.read().await;
    let adapter = state.get_central_adapter().await;
    let app = Arc::clone(&app_handle);
    if let Some(adapter) = adapter {
        let adapter = adapter.clone();
        let app_state = Arc::clone(&app_state);
        Ok(async_runtime::spawn(async move {
            let mut event_stream = adapter.events().await.unwrap();
            while let Some(event) = event_stream.next().await {
                match event {
                    btleplug::api::CentralEvent::DeviceDiscovered(id)
                    | btleplug::api::CentralEvent::DeviceUpdated(id) => {
                        let peripheral = adapter.peripheral(&id).await;
                        if let Ok(peripheral) = peripheral {
                            let properties = peripheral.properties().await.ok().flatten();
                            let state = app_state.write().await;
                            state.update_found_peripherals(id.clone()).await;
                            let item = PeripheralItem::new(&peripheral, properties.as_ref()).await;
                            app.emit("peripheral_found", item).unwrap();
                        }
                    }
                    btleplug::api::CentralEvent::DeviceConnected(id) => {
                        let state = app_state.write().await;
                        state.set_connected_peripheral(id).await;
                        app.emit("peripheral_connected", ()).unwrap();
                    }
                    btleplug::api::CentralEvent::DeviceDisconnected(_id) => {
                        let state = app_state.write().await;
                        state.clear_connected_peripheral().await;
                        app.emit("peripheral_disconnected", ()).unwrap();
                    }
                    _ => {}
                }
            }
        }))
    } else {
        app.emit("central_state_updated", ()).unwrap();
        Err(errors::AppError::BluetoothNotReady)
    }
}

pub async fn get_central_adapter() -> Result<Adapter, errors::AppError> {
    let manager = Manager::new()
        .await
        .map_err(|_| errors::AppError::BluetoothNotReady)?;
    let adapters = manager
        .adapters()
        .await
        .map_err(|_| errors::AppError::BluetoothAdapterNotFound)?;

    let mut found_adapter = None;
    for adpater in adapters {
        let state = adpater.adapter_state().await;
        if let Ok(state) = state {
            if state == CentralState::PoweredOn {
                found_adapter = Some(adpater);
                break;
            }
        }
    }

    found_adapter.ok_or(errors::AppError::NoAvailableBluetoothAdapter)
}

pub async fn start_scan(
    app_handle: Arc<AppHandle>,
    app_state: State<'_, Arc<RwLock<AppState>>>,
) -> Result<(), errors::AppError> {
    let state = app_state.read().await;
    let adapter = state.get_central_adapter().await;
    if let Some(adapter) = adapter {
        adapter
            .start_scan(ScanFilter::default())
            .await
            .map_err(|_| errors::AppError::UnableToStartScan)?;
        state.set_scanning(true).await;
        state.clear_found_peripherals().await;
        app_handle.emit("scanning_started", ()).unwrap();
        Ok(())
    } else {
        Err(errors::AppError::NoAvailableBluetoothAdapter)
    }
}

pub async fn stop_scan(
    app_handle: Arc<AppHandle>,
    app_state: State<'_, Arc<RwLock<AppState>>>,
) -> Result<(), errors::AppError> {
    let state = app_state.read().await;
    let adapter = state.get_central_adapter().await;
    if let Some(adapter) = adapter {
        adapter
            .stop_scan()
            .await
            .map_err(|_| errors::AppError::UnableToStopScan)?;
        state.set_scanning(false).await;
        app_handle.emit("scanning_stopped", ()).unwrap();
        Ok(())
    } else {
        Err(errors::AppError::NoAvailableBluetoothAdapter)
    }
}
